/**
 * @file   UtilitiesLib/ExecutionPath.h
 * @date   Aug 30, 2018
 * @author Bernd Doser, HITS gGmbH
 */

#pragma once

namespace pink {

/// Type for execution path
enum class ExecutionPath
{
    UNDEFINED,
    TRAIN,
    MAP
};

} // namespace pink
