/**
 * @file   Pink/main_gpu.h
 * @brief  Generic main routine of PINK
 * @date   Oct 15, 2018
 * @author Bernd Doser, HITS gGmbH
 */

#pragma once

#include "UtilitiesLib/InputData.h"

namespace pink {

void main_gpu(InputData const & input_data);

} // namespace pink
